#include "ScoreS1SQ.h"

ScoreS1SQ::ScoreS1SQ(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreS1SQ::ScoreS1SQ]  ";

  // read parameter file
  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile", &genoStream);
  dipGenoVec = new int[nsamples];

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new double[nsamples];

  if ( ioh.getBoolParam("ancprob") )  {
    errstream << "score requires integer-valued local ancestry";
    throw(Exception(errstream.str()));
  }

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  // ioh.dbgShowParams();

  return;
}

ScoreS1SQ::~ScoreS1SQ(void)  {

  if ( outputStream.is_open() )     {  outputStream.close();     }
  if ( genoStream.is_open()   )     {  genoStream.close();       }
  if ( phenoStream.is_open()  )     {  phenoStream.close();      }
  if ( localAncStream.is_open() )   {  localAncStream.close();  }

  // close file streams and free memory
  delete dipGenoVec;
  delete dipPhenoVec;
  delete dipLocalAnc;
}

int ScoreS1SQ::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreS1SQ::evaluate]  ";

  bool output_to_file = ioh.getBoolParam("output2file"); 

  try { ioh.readRealLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    try { ioh.readIntLine(&genoStream, nsamples, &dipGenoVec); }   
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading genotype" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double s;
    try  {
      s = evalS1SQ();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;
    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }
  }

  return 0;
}


double ScoreS1SQ::evalS1SQ(void)  {
  ostringstream errstream;
  errstream << "[ScoreS1SQ::evalS1SQ]  ";

  double sum;
  int count;

  int *baddata = new int[nsamples];
  double *gamma = new double[nsamples];   // genotype
  double *phi   = new double[nsamples];   // phenotype

  count = 0;
  for(int i=0;i<nsamples;i++)  {      // note:  assume phenotypes valid (?)
    baddata[i] = 0;
    if ( dipGenoVec[i] == 9 ) 
      baddata[i] = 1;
    if ( !ioh.getBoolParam("ancprob") )  {
      if ( dipLocalAnc[i] == 9 )  {
        baddata[i] = 1;
      }
    }
    if ( baddata[i] != 1 )
      count++;
  }

  // prevent division by zero
  if ( count == 0 )  { 
    errstream << "no valid data";
    throw(Exception(errstream.str()));
  }

  // mean-adjust genotype
  count = 0;
  sum = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    sum += dipGenoVec[i];
    count++;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    gamma[i] = dipGenoVec[i] - (sum/count);
  }

  // mean-adjust phenotype
  count = 0;
  sum = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    sum += dipPhenoVec[i];
    count++;
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] )
      continue;
    phi[i] = dipPhenoVec[i] - (sum/count);
  }

  double sumx = 0.0, sumy = 0.0, sumxx = 0.0, sumxy = 0.0, sumyy = 0.0;
  int sum1 = 0;

  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    if ( dipLocalAnc[i] != 0 ) 
      continue;
    sumx  += gamma[i];
    sumy  += phi[i];
    sumxx += (gamma[i]*gamma[i]);
    sumxy += (gamma[i]*phi[i]);
    sumyy += (phi[i]*phi[i]);
    sum1  += 1;
  }

  if ( sum1 == 0 )  {
    errstream << "no samples have zero localanc";
    throw(Exception(errstream.str()));
  }

  double mx = sumx/sum1;
  double my = sumy/sum1;
  double sx = sqrt(sumxx/sum1 - mx*mx);
  double sy = sqrt(sumyy/sum1 - my*my);
 
  if ( sx < 1e-8 || sy < 1e-8 )  {
    errstream << "zero variance";
    throw(Exception(errstream.str()));
  }
  double corrAA = (sumxy/sum1 - mx*my)/(sx*sy);
  int nAA = sum1;

  sumx = 0.0, sumy = 0.0, sumxx = 0.0, sumxy = 0.0, sumyy = 0.0;
  sum1 = 0;

  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) 
      continue;
    if ( dipLocalAnc[i] == 0 ) 
      continue;
    sumx  += gamma[i];
    sumy  += phi[i];
    sumxx += (gamma[i]*gamma[i]);
    sumxy += (gamma[i]*phi[i]);
    sumyy += (phi[i]*phi[i]);
    sum1  += 1;
  }

  if ( sum1 == 0 )  {
    errstream << "no samples have non-zero localanc";
    throw(Exception(errstream.str()));
  }

  mx = sumx/sum1;
  my = sumy/sum1;
  sx = sqrt(sumxx/sum1 - mx*mx);
  sy = sqrt(sumyy/sum1 - my*my);
 
  if ( sx < 1e-8 || sy < 1e-8 )  {
    errstream << "zero variance";
    throw(Exception(errstream.str()));
  }

  double corrAE = (sumxy/sum1 - mx*my)/(sx*sy);
  int nAE = sum1;
 
  delete baddata;
  delete gamma;
  delete phi;

  return( nAE*corrAE*corrAE + nAA*corrAA*corrAA ); 

} // end of eval
