#include "ScoreSAS.h"

double CinR_sas(double R, void *sasparams);
double CinpA_sas(double R, void *sasparams);
double CinpE_sas(double R, void *sasparams);

double fdCdR(double R, sasparams_t *p);
double fdCdpA(double R, sasparams_t *p);
double fdCdpE(double R, sasparams_t *p);

double C( sasparams_t *sasparams );
double dCdR( sasparams_t *sasparams );
double dCdpA( sasparams_t *sasparams );
double dCdpE( sasparams_t *sasparams );


double ScoreSAS::getSASEstimatedImputationR2(int *g, int *P, int *A, int nsamples, int anc)  {

  int n=0;
  double mean =0;
  double est_var = 0;
  double est_freq = 0;
  double est_allele_var =0;
  double rhat;

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || A[i] == 9 ) 
      continue;
    if(A[i] == anc ){
      mean += g[i];
      n++;
    }
  }
  est_freq = mean/(2*n);
  mean /= n;

  if(est_freq<0.01|| est_freq > 0.99){
    //fprintf(stderr,"potential overflow error in estimation of imputation error rate\n");
    return 1;
  }

  /*estimate variance of dosages*/
    est_allele_var = 2*(est_freq)*(1-est_freq);
  //est_allele_var = 2*(f)*(1-f);

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || A[i] == 9 || A[i] != anc) 
      continue;
    if(A[i] == anc ){
      double dose = (double)(g[i]);
      est_var += ((dose-mean)*(dose-mean));
    }
  }
  est_var /= (n-1);

  rhat = (est_var / (est_allele_var));
  if(rhat<0.1){
    rhat=0.1;
  }
  if(rhat>1){
    rhat=1;
  }

  //fprintf(stderr,"anc %d mean %f freq %f est_var %f est_allele_var %f rhat %f\n",anc,mean,est_freq,est_var, est_allele_var, rhat);

  return rhat;
}
ScoreSAS::ScoreSAS(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreSAS::ScoreSAS]  ";

  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile", &genoStream);
  dipGenoVec = new int[nsamples];

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new int[nsamples];

  globalAnc = new double[nsamples];

  genoDosages = new double[nsamples];
  ioh.getParam("genoformat",&genoform);
  use_adjustment = ioh.getBoolParam("adjustment"); 

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  if ( ioh.getBoolParam("oddsratio") )  {
    ioh.openOutput("orfile",&oddsratioStream);
  }

  return;
}

ScoreSAS::~ScoreSAS(void)  {
  string str;

  // close file streams 
  if ( outputStream.is_open() )     {  outputStream.close();     }
  if ( oddsratioStream.is_open() )  {  oddsratioStream.close();  }
  if ( genoStream.is_open()   )     {  genoStream.close();       }
  if ( phenoStream.is_open()  )     {  phenoStream.close();      }
  if ( localAncStream.is_open() )   {  localAncStream.close();   }
  // mem: global anc stream already closed

  // free memory
  delete dipGenoVec;
  delete dipPhenoVec;
  delete globalAnc;
  //delete mysasparams->TS;
  delete genoDosages;

  delete dipLocalAnc;
    //delete mysasparams->AS;
  delete mysasparams;

}

int ScoreSAS::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreSAS::evaluate]  ";

  bool output_to_file = ioh.getBoolParam("output2file"); 
  bool output_oddsratio = ioh.getBoolParam("oddsratio");

  try { ioh.readIntLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << "error reading phenotype" << endl;
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  str = ioh.getStringParam("thetafile");
  try { ioh.readOneColumnFile(str, nsamples, &globalAnc); }
  catch ( Exception& e )  {
    errstream << "error reading globalanc" << endl;
    errstream << e.getMessage();
    throw( Exception (errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    if( !genoform.compare("dosage") ){
      geno_is_dosage = true;
      try { ioh.readRealLine(&genoStream, nsamples, &genoDosages); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
    }
    else if ( !genoform.compare("prob") )  {
      geno_is_dosage = true;
      double *tempVec = new double[2*nsamples];
      try { ioh.readRealLine(&genoStream, 2*nsamples, &tempVec); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
      for(int i=0;i<nsamples;i++)  {
        double p2 = tempVec[2*i];
        double p1 = tempVec[2*i+1];
        genoDosages[i] = 2.0*p2 + p1;
      }
      delete tempVec;
    }
    else  {    // genoformat is "integer"
      geno_is_dosage = false;
      try { ioh.readIntLine(&genoStream, nsamples, &dipGenoVec); }
      catch ( Exception& e )  {
        errstream << "SNP # " << snpcount << endl;
        errstream << "error reading genotype" << endl;
        errstream << e.getMessage();
        throw( Exception(errstream.str()) );
      }
    }


    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double s;
    try  {
      s = evalSAS();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }

    if ( output_oddsratio )  {
      oddsratioStream << fixed << setw(10) << 
         oddsratio << endl;
    }
  }

  return 0;
}




#define RMIN 0.1
#define RMAX 10.0
#define PAMIN 0.0001
#define PAMAX 0.9999
#define PEMIN 0.0001
#define PEMAX 0.9999

#define MAXBIGITER 100
#define MAXSMALLITER 40
#define CONVERGE 1e-6
#define SMALL 1e-4
#define DBG 0


double ScoreSAS::evalSAS()  {
  ostringstream errstream;
  errstream << "[ScoreSAS::evalSAS]  ";

  const gsl_min_fminimizer_type *T0, *T1, *T2;
  gsl_min_fminimizer *s0, *s1, *s2;
  gsl_function F0, F1, F2;

  int done = 0, bigiter = 0, smalliter = 0, status, iter;
  double nullscore, causalscore;

  /* store initial values of parameters and compute null score */
  string str;

  if(geno_is_dosage){
    /*round the dosages*/
    for(int i =0; i < nsamples; i++){
      if(genoDosages[i]<0.5)  {
	  dipGenoVec[i] = 0;
      }
      else  {
        if(genoDosages[i]>1.5)  {
          dipGenoVec[i] = 2;
        } 
        else  {
          dipGenoVec[i] = 1;
        }
      }
    }
  }  
  init_sasparams(dipGenoVec, dipPhenoVec, dipLocalAnc, globalAnc); 


  /* Set up to use GSL minimization routine */

  T0 = gsl_min_fminimizer_brent;
  T1 = gsl_min_fminimizer_brent;
  T2 = gsl_min_fminimizer_brent;

  s0 = gsl_min_fminimizer_alloc(T0);
  s1 = gsl_min_fminimizer_alloc(T1);
  s2 = gsl_min_fminimizer_alloc(T2);

  F0.function = &CinR_sas;
  F1.function = &CinpA_sas;
  F2.function = &CinpE_sas;

  F0.params = mysasparams;
  F1.params = mysasparams;
  F2.params = mysasparams;

  gsl_set_error_handler_off();

  for(iter=0;iter<2;iter++)  {

    done = 0;
    while ( !done )  {

      double newR, newpA, newpE;
      double oldR = mysasparams->R, oldpA = mysasparams->pA, oldpE = mysasparams->pE;

      if ( iter == 1 )  {
        /*  Minimize over R - causal score only */
        mysasparams->R = oldR;
        if ( fdCdR(RMIN,mysasparams)*fdCdR(RMAX,mysasparams) > 0.0 )  {
          double c0 = CinR_sas(RMIN,mysasparams), c1 = CinR_sas(RMAX,mysasparams);
          newR = ( c0 < c1 ? RMIN : RMAX );
        }
        else {
          if ( (status = gsl_min_fminimizer_set(s0,&F0,oldR,RMIN,RMAX)) != 0 )  {
  	    try { status = binsearch(&fdCdR,mysasparams,RMIN,RMAX,&newR);  }  
            catch ( Exception e )  {
              errstream << "(causal score) ";
              errstream << "binsearch did not find minimum in R" << endl;
              errstream << errorString();
              errstream << e.getMessage();
              throw(Exception(errstream.str()));
	    }
          }
          else  {
            smalliter = 0;
            do  {

              if ( (status = gsl_min_fminimizer_iterate(s0)) != 0 )  {
                errstream << "(causal score) ";
                errstream << "GSL iteration did not find minimum in R";
                errstream << endl << gsl_strerror(status);
                errstream << errorString();
                throw(Exception(errstream.str()));
	      }

	      double a = gsl_min_fminimizer_x_lower(s0);
	      double b = gsl_min_fminimizer_x_upper(s0);
	      newR = gsl_min_fminimizer_x_minimum(s0);
	      status = gsl_min_test_interval(a,b,CONVERGE,0.0);
	    } while ( status == GSL_CONTINUE && smalliter++ < MAXSMALLITER );
          }
        }
        mysasparams->R = newR;

        if ( DBG )  {
          printf("  oldR = %f  newR = %f\n", oldR, newR);
        }

      }  /* end minimization of R */


      /*  Minimize over pA */
      mysasparams->pA = oldpA;

      /* Question:  What special cases need to be handled? */
      /* For semi-monomorphic SNPs?                        */
      /* E.g., to prevent division by zero                 */

      if ( fdCdpA(PAMIN,mysasparams)*fdCdpA(PAMAX,mysasparams) > 0.0 )  {
        double c0 = CinpA_sas(PAMIN,mysasparams), c1 = CinpA_sas(PAMAX,mysasparams);
        newpA = ( c0 < c1 ? PAMIN : PAMAX );
      }
      else {
        if ( (status = gsl_min_fminimizer_set(s1,&F1,oldpA,PAMIN,PAMAX)) != 0 )  {
  	  try { status = binsearch(&fdCdpA,mysasparams,PAMIN,PAMAX,&newpA); }  
          catch ( Exception& e )  {
            errstream << ( iter == 1 ? "(causal score)  " : "(null score) " );
            errstream << "binsearch did not find minimum in pA" << endl;
            errstream << errorString();
            errstream << e.getMessage();
            throw(Exception(errstream.str()));
	  }
        }
        else  {
          smalliter = 0;
          do  {
            if ( (status = gsl_min_fminimizer_iterate(s1)) != 0 )  {
              errstream << ( iter == 1 ? "(causal score)  " : "(null score) " );
              errstream << "GSL iteration did not find minimum in pA";
              errstream << endl << gsl_strerror(status);
              errstream << errorString();
              throw(Exception(errstream.str()));
	    }
	    double a = gsl_min_fminimizer_x_lower(s1);
	    double b = gsl_min_fminimizer_x_upper(s1);
	    newpA = gsl_min_fminimizer_x_minimum(s1);

	    status = gsl_min_test_interval(a,b,CONVERGE,0.0);
	  } while ( status == GSL_CONTINUE && smalliter++ < MAXSMALLITER );
        }
      }
      mysasparams->pA = newpA;

      if ( DBG )  {
        printf("  oldpA = %f  newpA = %f\n", oldpA, newpA);
      }

      /*  Minimize over pE */
      mysasparams->pE = oldpE;
      if ( fdCdpE(PEMIN,mysasparams)*fdCdpE(PEMAX,mysasparams) > 0.0 )  {
        double c0 = CinpE_sas(PEMIN,mysasparams), c1 = CinpE_sas(PEMAX,mysasparams);
        newpE = ( c0 < c1 ? PEMIN : PEMAX );
      }
      else {
        if ( (status = gsl_min_fminimizer_set(s2,&F2,oldpE,PEMIN,PEMAX)) != 0 )  {
  	  try { status = binsearch(&fdCdpE,mysasparams,PEMIN,PEMAX,&newpE); }  
          catch ( Exception& e )  {
            errstream << ( iter == 1 ? "(causal score)  " : "(null score)  ");
            errstream << "binsearch did not find minimum in pE" << endl;
            errstream << errorString();
            errstream << e.getMessage();
            throw(Exception(errstream.str()));
	  }
        }
        else  {
          smalliter = 0;
          do  {
            if ( (status = gsl_min_fminimizer_iterate(s2)) != 0 )  {
              errstream << ( iter == 1 ? "(causal score)  " : "(null score)  ");
              errstream << "GSL iteration did not find minimum in pE";
              errstream << endl << gsl_strerror(status);
              errstream << errorString();
              throw( Exception(errstream.str()));
	    }
	    double a = gsl_min_fminimizer_x_lower(s2);
	    double b = gsl_min_fminimizer_x_upper(s2);
	    newpE = gsl_min_fminimizer_x_minimum(s2);
	    status = gsl_min_test_interval(a,b,CONVERGE,0.0);
	  } while ( status == GSL_CONTINUE && smalliter++ < MAXSMALLITER );
        }
      }
      mysasparams->pE = newpE;

      if ( DBG )  {
        printf("  oldpE = %f  newpE = %f\n", oldpE, newpE);
      }


      if ( iter == 1 )  {
        if ( fabs(newR-oldR) < SMALL && fabs(newpA-oldpA) < 
               SMALL && fabs(newpE-oldpE) < SMALL )  {
          done = 1;
        }
      }
      else  {
        if ( fabs(newpA-oldpA) < SMALL && fabs(newpE-oldpE) < SMALL )  {
          done = 1;
        }
      }

      if ( DBG )  {
        printf("delta = (%f,%f,%f)\n\n", fabs(newR-oldR), fabs(newpA-oldpA),
          fabs(newpE-oldpE));
        cout << "causal score " << -C(mysasparams) << endl;
      }

      if ( bigiter++ > MAXBIGITER )  {
        done = 1;
      }

    }  /* end main loop */


    if ( bigiter > MAXBIGITER )  {
      errstream << ( iter == 1 ? "(causal score)  " : "(null score)  ");
      errstream << "main iteration did not converge"<<endl;
      errstream << errorString();
      throw(Exception(errstream.str()));
    }

    if ( iter == 0 )  {
      nullscore = -C(mysasparams);
    }
    else  {
      causalscore = -C(mysasparams);
      setOddsRatio(mysasparams->R);
    }

    if ( DBG )  {
      if ( iter == 0 )  {
        cout << "  null score " << nullscore << endl;
      }
      else  {
        cout << "  null score " << nullscore ;
        cout << "  causal score " << causalscore << endl;
      }
    }

  }  /* end of double loop */

  if ( causalscore < nullscore )  {
    errstream << "causal score " << 
      fixed << setw(10) << setprecision(4) << causalscore << 
      " is less than null score " << 
      fixed << setw(10) << setprecision(4) << nullscore << 
      endl;
    errstream << errorString();
    throw( Exception(errstream.str()) ); /*added by Bogdan Thu July 5th; bugfix 1.2 -> 1.3*/
  }

  delete(mysasparams->AS);
  delete(mysasparams->TS);
  return ( causalscore-nullscore );

}


/* ------------------------------------------------------------------- */
void ScoreSAS::init_sasparams(int *G, int *P, int *A, double *T)  {

  sasparams_t *p = new sasparams_t;

  p->XAARRT = p->XAARVT = p->XAAVVT = p->XAARRS = p->XAARVS = p->XAAVVS =
  p->XEERRT = p->XEERVT = p->XEEVVT = p->XEERRS = p->XEERVS = p->XEEVVS =
  p->XAERRT = p->XAERVT = p->XAEVVT = p->XAERRS = p->XAERVS = p->XAEVVS = 0.0;

  p->XAVT = p->XART = p->XAVS = p->XARS = 0.0;
  p->XEVT = p->XERT = p->XEVS = p->XERS = 0.0;

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || G[i] == 9 || A[i] == 9 ) 
      continue;

    int g = G[i];
    int a = A[i];

    if ( P[i] == 0 )  {              /* Controls */
      if ( g == 0 )  {               /* VV */
        if ( a == 0 )  {             /* AA */
	  p->XAAVVT += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAEVVT += 1;
	}
	else  {                      /* EE */
	  p->XEEVVT += 1;
	}
      }
      else if ( g == 1 )  {          /* RV      */
        if ( a == 0 )  {             /* AA */
	  p->XAARVT += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERVT += 1;
	}
	else  {                      /* EE */
	  p->XEERVT += 1;
	}
      }
      else  {                        /* RR      */
        if ( a == 0 )  {             /* AA */
	  p->XAARRT += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERRT += 1;
	}
	else  {                      /* EE */
	  p->XEERRT += 1;
	}
      }
    }
    else  {                          /* Cases    */
      if ( g == 0 )  {               /* VV      */
        if ( a == 0 )  {             /* AA */
	  p->XAAVVS += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAEVVS += 1;
	}
	else  {                      /* EE */
	  p->XEEVVS += 1;
	}
      }
      else if ( g == 1 )  {          /* RV      */
        if ( a == 0 )  {             /* AA */
	  p->XAARVS += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERVS += 1;
	}
	else  {                      /* EE */
	  p->XEERVS += 1;
	}
      }
      else  {                        /* RR      */
        if ( a == 0 )  {             /* AA */
	  p->XAARRS += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERRS += 1;
	}
	else  {                      /* EE */
	  p->XEERRS += 1;
	}
      }
    }
  }


  /* pack Cases into arrays */
  p->AS = new int[nsamples];
  p->TS = new double[nsamples]; 

  p->nt = 0;
  p->local_anc_is_prob = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 0 || P[i] == 9 || A[i] == 9 || G[i] == 9)  
      continue;
    
    p->TS[p->nt] = T[i];
    p->AS[p->nt] = A[i];
    p->nt++;
  }


  p->XART = (2*p->XAARRT + p->XAARVT) + 0.5 * (2*p->XAERRT + p->XAERVT);
  p->XARS = (2*p->XAARRS + p->XAARVS) + 0.5 * (2*p->XAERRS + p->XAERVS);
  p->XERT = (2*p->XEERRT + p->XEERVT) + 0.5 * (2*p->XAERRT + p->XAERVT);
  p->XERS = (2*p->XEERRS + p->XEERVS) + 0.5 * (2*p->XAERRS + p->XAERVS);


  p->XAVT = (2*p->XAAVVT + p->XAARVT) + 0.5  * (2*p->XAEVVT + p->XAERVT);
  p->XAVS = (2*p->XAAVVS + p->XAARVS) + 0.5 * (2*p->XAEVVS + p->XAERVS);
  p->XEVT = (2*p->XEEVVT + p->XEERVT) + 0.5 * (2*p->XAEVVT + p->XAERVT);
  p->XEVS = (2*p->XEEVVS + p->XEERVS) + 0.5 * (2*p->XAEVVS + p->XAERVS);

  for(int x =0; x < 1; x++){
    //fprintf(stderr,"%f %f %f %f %f %f %f %f\n",p->XART,p->XARS,p->XERT,p->XERS,p->XAVT,p->XAVS,p->XEVT,p->XEVS);
    //bugfix 1.1 to 1.2
    double estpa =  SMALL + p->XART/(p->XART+p->XAVT);
    double estpe =  SMALL + p->XERT/(p->XERT+p->XEVT);
    double estcpa = SMALL + p->XARS/(p->XARS+p->XAVS);
    double estcpe = SMALL + p->XERS/(p->XERS+p->XEVS);
    if(estpa>1) estpa = 1-SMALL;
    if(estpe>1) estpe = 1-SMALL;
    if(estcpa>1) estcpa = 1-SMALL;
    if(estcpe>1) estcpe = 1-SMALL;
    if(estpa<0) estpa = SMALL;
    if(estpe<0) estpe = SMALL;
    if(estcpa<0) estcpa = SMALL;
    if(estcpe<0) estcpe = SMALL;

    //fprintf(stderr,"%f %f %f %f (%f %f)\n",estpa,estpe,estcpa,estcpe, estpa/(estpa+estpe),estcpa/(estcpa+estcpe));
    
    p->XART = (2*p->XAARRT + p->XAARVT) + (estpa/(estpa+estpe)) * (2*p->XAERRT + p->XAERVT);
    p->XARS = (2*p->XAARRS + p->XAARVS) + (estcpa/(estcpa+estcpe)) * (2*p->XAERRS + p->XAERVS);
    p->XERT = (2*p->XEERRT + p->XEERVT) + (estpe/(estpa+estpe)) * (2*p->XAERRT + p->XAERVT);
    p->XERS = (2*p->XEERRS + p->XEERVS) + (estcpe/(estcpa+estcpe)) * (2*p->XAERRS + p->XAERVS);
    
    
    p->XAVT = (2*p->XAAVVT + p->XAARVT) + ((1-estpa)/(2-estpa-estpe))  * (2*p->XAEVVT + p->XAERVT);
    p->XAVS = (2*p->XAAVVS + p->XAARVS) + ((1-estcpa)/(2-estcpa-estcpe)) * (2*p->XAEVVS + p->XAERVS);
    p->XEVT = (2*p->XEEVVT + p->XEERVT) + ((1-estpe)/(2-estpa-estpe)) * (2*p->XAEVVT + p->XAERVT);
    p->XEVS = (2*p->XEEVVS + p->XEERVS) + ((1-estcpe)/(2-estcpa-estcpe)) * (2*p->XAEVVS + p->XAERVS);
    
    //fprintf(stderr,"%f %f %f %f %f %f %f %f\n",p->XART,p->XARS,p->XERT,p->XERS,p->XAVT,p->XAVS,p->XEVT,p->XEVS);
  }
  //bugfix 1.1 to 1.2
#define MINNR 0.1
  p->XEVT += MINNR;
  p->XEVS += MINNR;
  p->XAVT += MINNR;
  p->XAVS += MINNR;
  p->XERT += MINNR;
  p->XERS += MINNR;
  p->XART += MINNR;
  p->XARS += MINNR;

  //fprintf(stderr,"%f %f %f %f %f %f %f %f\n",p->XART,p->XARS,p->XERT,p->XERS,p->XAVT,p->XAVS,p->XEVT,p->XEVS);
  double numE = (  p->XERT );
  double denE = (  p->XERT + p->XEVT );
  
  double numA = (  p->XART );
  double denA = (  p->XART + p->XAVT );

  p->impPa = ( p->XARS + p->XART) / (p->XARS+p->XART+p->XAVS+p->XAVT);
  p->impPe = ( p->XERS + p->XERT) / (p->XERS+p->XERT+p->XEVS+p->XEVT);


  p->rhatA = 1;
  p->rhatE = 1;
  if(use_adjustment){
    p->rhatA = getSASEstimatedImputationR2(G, P, A, nsamples, 0);
    p->rhatE = getSASEstimatedImputationR2(G, P, A, nsamples, 2);
  }


  p->R = 1.0;
  p->pE = numE/denE;
  p->pA = numA/denA;

  mysasparams = p;
}

/* ------------------------------------------------------------------- */
double CinR_sas(double R, void *sasparams)  {
  sasparams_t *p = (sasparams_t *)sasparams;
  p->R = R;
  return C(p);
}

/* ------------------------------------------------------------------- */
double CinpA_sas(double pA, void *sasparams)  {
  sasparams_t *p = (sasparams_t *)sasparams;
  p->pA = pA;
  return C(p);
}

/* ------------------------------------------------------------------- */
double CinpE_sas(double pE, void *sasparams)  {
  sasparams_t *p = (sasparams_t *)sasparams;
  p->pE = pE;
  return C(p);
}

/* ------------------------------------------------------------------- */
double fdCdR(double R, sasparams_t *p)  {
  p->R = R;
  return(dCdR(p));
}

double fdCdpA(double pA, sasparams_t *p)  {
  p->pA = pA;
  return (dCdpA(p)); 
}

double fdCdpE(double pE, sasparams_t *p)  {
  p->pE = pE;
  return (dCdpE(p)); 
}

/* ------------------------------------------------------------------- */
double C(sasparams_t *p)  {

  int nt =  p->nt;
  /*
  double XAARRT = p->XAARRT;
  double XAARVT = p->XAARVT;
  double XAAVVT = p->XAAVVT;

  double XAARRS = p->XAARRS;
  double XAARVS = p->XAARVS;
  double XAAVVS = p->XAAVVS;

  double XEERRT = p->XEERRT;
  double XEERVT = p->XEERVT;
  double XEEVVT = p->XEEVVT;

  double XEERRS = p->XEERRS;
  double XEERVS = p->XEERVS;
  double XEEVVS = p->XEEVVS;

  double XAERRT = p->XAERRT;
  double XAERVT = p->XAERVT;
  double XAEVVT = p->XAEVVT;

  double XAERRS = p->XAERRS;
  double XAERVS = p->XAERVS;
  double XAEVVS = p->XAEVVS;
  */
  double R = p->R;
  double pA = p->pA;
  double pE = p->pE;

  /*
  double Ra = (1+(R-1)*(p->rhatA + p->impPa*(1-p->rhatA)))
    / (1+(R-1)*p->impPa*(1-p->rhatA));
  double Re = (1+(R-1)*(p->rhatE + p->impPe*(1-p->rhatE)))
    / (1+(R-1)*p->impPe*(1-p->rhatE));
  */

  double Ra = (1+(R-1)*(p->rhatA + pA*(1-p->rhatA)))
    / (1+(R-1)*pA*(1-p->rhatA));
  double Re = (1+(R-1)*(p->rhatE + p->pE*(1-p->rhatE)))
    / (1+(R-1)*pE*(1-p->rhatE));



  double *A0, *A1, *A2;
  int *AS;
  if ( p->local_anc_is_prob )  {
    A0 = p->AS0;
    A1 = p->AS1;
    A2 = p->AS2;
  }
  else  {
    AS = p->AS;
  }

  double *T = p->TS;

  double casepA, casepE, result, resultD, sum, rho;
  int i;

  casepA = Ra*pA / (1.0 + (Ra-1.0)*pA);
  casepE = Re*pE / (1.0 + (Re-1.0)*pE);
  result = 0.0;
  resultD = 0.0;

  result += p->XERT *log(pE) + p->XEVT *log(1-pE) + p->XART * log (pA) + p->XAVT *log(1-pA);
  result += p->XERS *log(casepE) + p->XEVS *log(1-casepE) + p->XARS * log (casepA) + p->XAVS *log(1-casepA);

  /*
  result += ( 2*XEERRT + XEERVT + XAERRT ) * log(pE)
    + ( XEERVT + 2*XEEVVT + XAEVVT ) * log(1-pE) 
    + ( 2*XAARRT + XAARVT + XAERRT ) * log(pA)
    + ( XAARVT + 2*XAAVVT + XAEVVT ) * log(1-pA) 
    +  XAERVT * log( pA + pE - 2*pA*pE );

  result += ( 2*XEERRS + XEERVS + XAERRS ) * log(casepE)
    + ( XEERVS + 2*XEEVVS + XAEVVS ) * log(1-casepE) 
    + ( 2*XAARRS + XAARVS + XAERRS ) * log(casepA)
    + ( XAARVS + 2*XAAVVS + XAEVVS ) * log(1-casepA) 
    +  XAERVS * log( casepA + casepE - 2*casepA*casepE );
  */

  sum = 0.0;
  rho = ( pE*R + (1-pE) )/( pA*R + (1-pA) );
  for(i=0;i<nt;i++)  {
    double p2 = rho*rho*T[i]*T[i];
    double p1 = 2*rho*T[i]*(1-T[i]);
    double p0 = (1-T[i])*(1-T[i]);
    double psum = p0+p1+p2;
    if ( p->local_anc_is_prob )  {
      if ( A2[i] > 1e-8 )  
        sum += A2[i]*log( p2/psum );
      if ( A1[i] > 1e-8 ) 
        sum += A1[i]*log( p1/psum );
      if ( A0[i] > 1e-8 )
        sum += A0[i]*log( p0/psum );
    }
    else  {
      if ( AS[i] == 2 )  
        sum += log( p2/psum );
      else if ( AS[i] == 1 ) 
        sum += log( p1/psum );
      if ( AS[i] == 0 )
        sum += log( p0/psum );
    }
  }
  result += sum;

  return ( -(2*result) );    /* maximize by minimizing negative */

}



/* ------------------------------------------------------------------- */
double dCdR(sasparams_t *p)  {

  int nt  = p->nt;

  /* double XAARRS = p->XAARRS;
  double XAARVS = p->XAARVS;
  double XAAVVS = p->XAAVVS;

  double XEERRS = p->XEERRS;
  double XEERVS = p->XEERVS;
  double XEEVVS = p->XEEVVS;

  double XAERRS = p->XAERRS;
  double XAERVS = p->XAERVS;
  double XAEVVS = p->XAEVVS;*/

  double R = p->R;
  double pA = p->pA;
  double pE = p->pE;
  double *T = p->TS;

  /*  
     double Ra = (1+(R-1)*(p->rhatA + p->impPa*(1-p->rhatA)))
    / (1+(R-1)*p->impPa*(1-p->rhatA));
  double Re = (1+(R-1)*(p->rhatE + p->impPe*(1-p->rhatE)))
    / (1+(R-1)*p->impPe*(1-p->rhatE));
  */
  double Ra = (1+(R-1)*(p->rhatA + pA*(1-p->rhatA)))
    / (1+(R-1)*pA*(1-p->rhatA));
  double Re = (1+(R-1)*(p->rhatE + p->pE*(1-p->rhatE)))
    / (1+(R-1)*pE*(1-p->rhatE));


  double dpRa =   (p->rhatA) / 
    pow((p->rhatA + p->impPa*(1-p->rhatA)),2);
  double dpRe =   (p->rhatE) / 
    pow((p->rhatE + p->impPe*(1-p->rhatE)),2);



  double result = 
    ( 1/(1+(Ra-1)*pA) ) * ( p->XARS*(1-pA)/Ra + p->XAVS *(-pA))*dpRa +
    ( 1/(1+(Re-1)*pE) ) * ( p->XERS*(1-pE)/Re + p->XEVS *(-pE))*dpRe;
    /*
    ( 2*XEERRS + XEERVS + XAERRS + 2*XAARRS + XAARVS + XAERRS + XAERVS ) / R + 
    ( 2*XEERRS + XEERVS + XAERRS + XEERVS + 2*XEEVVS + XAEVVS + XAERVS ) * 
         ( -pE/(1 + (R-1)*pE) ) +
    ( 2*XAARRS + XAARVS + XAERRS + XAARVS + 2*XAAVVS + XAEVVS + XAERVS )  *
         ( -pA/(1 + (R-1)*pA) );
    */
  /* drho/rho is (pE-pA)/(1-pA+R*pA)*(1-pE + R*pE)  */

  double rho = ( pE*R + (1-pE) )/( pA*R + (1-pA) );
  double drho = (pE-pA)/((1-pA+R*pA)*(1-pA+R*pA));

  //  double nee = (XEERRS + XEERVS + XEEVVS);
  // double nae = (XAERRS + XAERVS + XAEVVS);
  // result += (2*nee+nae)*drho/rho;



  double sum = 0.0;
  int i;
  for(i=0;i<nt;i++)  {
    double b = rho*rho*T[i]*T[i] + 2*rho*T[i]*(1-T[i]) + (1-T[i])*(1-T[i]);
    double db = 2*rho*drho*T[i]*T[i] + 2*drho*T[i]*(1-T[i]);
    sum += (db/b);
  }
  result -= sum;

  return (2*result);
}


/* ------------------------------------------------------------------- */
double dCdpA(sasparams_t *p)  {
  int nt = p->nt;

  /*
    double XAARRT = p->XAARRT;
    double XAARVT = p->XAARVT;
    double XAAVVT = p->XAAVVT;
    
    double XAARRS = p->XAARRS;
    double XAARVS = p->XAARVS;
    double XAAVVS = p->XAAVVS;
    

    double XEERRS = p->XEERRS;
    double XEERVS = p->XEERVS;
    double XEEVVS = p->XEEVVS;
    
    double XAERRT = p->XAERRT;
    double XAERVT = p->XAERVT;
  double XAEVVT = p->XAEVVT;

  double XAERRS = p->XAERRS;
  double XAERVS = p->XAERVS;
  double XAEVVS = p->XAEVVS;
  */

  double R = p->R;
  double pA = p->pA;
  double pE = p->pE;
  double *T = p->TS;
  /*  double Ra = (1+(R-1)*(p->rhatA + p->impPa*(1-p->rhatA)))
    / (1+(R-1)*p->impPa*(1-p->rhatA));
  */
  double Ra = (1+(R-1)*(p->rhatA + pA*(1-p->rhatA)))
    / (1+(R-1)*pA*(1-p->rhatA));
  

  double result = 
    p->XART / pA - p->XAVT /(1-pA) +
    (1/(1+(Ra-1)*pA)) * ( p->XARS / pA - p-> XAVS*Ra /(1-pA));

    /*    ( 2*XAARRT + XAARVT + XAERRT + 2*XAARRS + XAARVS + XAERRS ) / pA - 
    ( XAARVT + 2*XAAVVT + XAEVVT  + XAARVS + 2*XAAVVS + XAEVVS ) / (1-pA) + 
    ( 2*XAARRS + XAARVS + XAERRS + XAARVS + 2*XAAVVS + XAEVVS + XAERVS ) * 
      (1-R)/(1 + (R-1)*pA ) +
    ( XAERVT + XAERVS ) * (1-2*pE)/(pE + pA - 2*pE*pA);
    */
  double rho = ( pE*R + (1-pE) )/( pA*R + (1-pA) );
  double drho = -(R-1)*(1-pE+R*pE)/( (1-pA+R*pA)*(1-pA+R*pA) );

  // double nee = (XEERRS + XEERVS + XEEVVS);
  //double nae = (XAERRS + XAERVS + XAEVVS);
  //result += (2*nee+nae)*drho/rho;

  double sum = 0.0;
  int i;
  for(i=0;i<nt;i++)  {
    double b = rho*rho*T[i]*T[i] + 2*rho*T[i]*(1-T[i]) + (1-T[i])*(1-T[i]);
    double db = 2*rho*drho*T[i]*T[i] + 2*drho*T[i]*(1-T[i]);
    sum += (db/b);
  }
  result -= sum;

  return (2*result);

}

/* ------------------------------------------------------------------- */
double dCdpE(sasparams_t *p)  {
  int nt = p->nt;

  /*double XEERRT = p->XEERRT;
  double XEERVT = p->XEERVT;
  double XEEVVT = p->XEEVVT;

  double XEERRS = p->XEERRS;
  double XEERVS = p->XEERVS;
  double XEEVVS = p->XEEVVS;

  double XAERRT = p->XAERRT;
  double XAERVT = p->XAERVT;
  double XAEVVT = p->XAEVVT;

  double XAERRS = p->XAERRS;
  double XAERVS = p->XAERVS;
  double XAEVVS = p->XAEVVS;
  */
  double R = p->R;
  double pA = p->pA;
  double pE = p->pE;
  double *T = p->TS;

  /*  double Re = (1+(R-1)*(p->rhatE + p->impPe*(1-p->rhatE)))
    / (1+(R-1)*p->impPe*(1-p->rhatE));
  */
  double Re = (1+(R-1)*(p->rhatE + p->pE*(1-p->rhatE)))
    / (1+(R-1)*pE*(1-p->rhatE));

  double result = 
    p->XERT / pE - p->XAVT /(1-pE) +
    (1/(1+(Re-1)*pE)) * ( p->XERS / pE - p-> XEVS*Re /(1-pE));

    /*    ( 2*XEERRT + XEERVT + XAERRT + 2*XEERRS + XEERVS + XAERRS ) / pE - 
    ( XEERVT + 2*XEEVVT + XAEVVT  + XEERVS + 2*XEEVVS + XAEVVS ) / (1-pE) + 
    ( 2*XEERRS + XEERVS + XAERRS + XEERVS + 2*XEEVVS + XAEVVS + XAERVS ) * 
      (1-R)/(1 + (R-1)*pE ) +
    ( XAERVT + XAERVS ) * (1-2*pA)/(pE + pA - 2*pE*pA);
    */

  /* instead of drho/rho, evaluate (R-1)/(pE*R + (1-pE))  */

  double rho = ( pE*R + (1-pE) )/( pA*R + (1-pA) );
  double drho = (R-1)/( 1-pA+R*pA );

  //  double nee = (XEERRS + XEERVS + XEEVVS);
  // double nae = (XAERRS + XAERVS + XAEVVS);
  // result += (2*nee+nae)*drho/rho;

  double sum = 0.0;
  int i;
  for(i=0;i<nt;i++)  {
    double b = rho*rho*T[i]*T[i] + 2*rho*T[i]*(1-T[i]) + (1-T[i])*(1-T[i]);
    double db = 2*rho*drho*T[i]*T[i] + 2*drho*T[i]*(1-T[i]);
    sum += (db/b);
  }
  result -= sum;

  return (2*result);

}


/* ------------------------------------------------------------------- */
#define BINSEARCH_MAXITER 100
int ScoreSAS::binsearch(double(*f)(double, sasparams_t *),
  sasparams_t *p, double xmin, double xmax, double *xout)  {
  ostringstream errstream;
  errstream << "[ScoreSAS::binsearch]  ";

  double xleft  = xmin;
  double xright = xmax;
  double fleft  = f(xleft, p);
  double fright = f(xright, p);
  int itcount = 0;

  if ( fleft*fright > 0 )  {
    errstream << "no minimum on search interval";
    throw( Exception(errstream.str()) );
  }

  while ( (xright-xleft) > SMALL )  {

    double xmid = 0.5*(xleft+xright);
    double fmid = f(xmid, p);

    if ( fmid*fleft < 0 )  {
      xright = xmid;
      fright = fmid;
    }
    else if ( fmid*fright < 0 )  {
      xleft = xmid;
      fleft = fmid;
    }
    else  {
      return -1;
    }

    if ( itcount++ > BINSEARCH_MAXITER )  {
      errstream << "did not converge in " << (itcount-1) << "iterations";
      throw( Exception(errstream.str()) );
    }
  }

  *xout = 0.5*(xleft+xright);
  return 0;
}




/* ------------------------------------------------------------------- */
string ScoreSAS::errorString(void)  {
  ostringstream ss;
  ss << "R = " << fixed << setw(14) << setprecision(4) << mysasparams->R;
  ss << "  pA = " << fixed << setw(14) << setprecision(4) << mysasparams->pA;
  ss << "  pE = " << fixed << setw(14) << setprecision(4) << mysasparams->pE;
  ss << endl;
  ss << "Controls: " << endl;
  ss << " AR = "<< fixed << setw(14) << setprecision(4) << mysasparams->XART ;
  ss << " ER = "<< fixed << setw(14) << setprecision(4) << mysasparams->XERT ;
  ss << " AV = "<< fixed << setw(14) << setprecision(4) << mysasparams->XAVT ;
  ss << " EV = "<< fixed << setw(14) << setprecision(4) << mysasparams->XEVT ;
  ss <<endl;
  ss << "Controls: " << endl;
  ss << " AR = "<< fixed << setw(14) << setprecision(4) << mysasparams->XARS ;
  ss << " ER = "<< fixed << setw(14) << setprecision(4) << mysasparams->XERS ;
  ss << " AV = "<< fixed << setw(14) << setprecision(4) << mysasparams->XAVS ;
  ss << " EV = "<< fixed << setw(14) << setprecision(4) << mysasparams->XEVS ;
  ss <<endl;
  ss << "Imputation: " << endl;

  ss<< " RhatA = "<< fixed << setw(14) << setprecision(4) << mysasparams->rhatA ;
  ss<< " RhatE = "<< fixed << setw(14) << setprecision(4) << mysasparams->rhatE ;
  ss<< " fA = "<< fixed << setw(14) << setprecision(4) << mysasparams->impPa ;
  ss<< " fE = "<< fixed << setw(14) << setprecision(4) << mysasparams->impPe ;
  ss<<endl;

  /*

  ss << "  AARR = " << fixed << setw(14) << setprecision(4) << mysasparams->XAARRT;
  ss << "  AERR = " << fixed << setw(14) << setprecision(4) << mysasparams->XAERRT;
  ss << "  EERR = " << fixed << setw(14) << setprecision(4) << mysasparams->XEERRT;
  ss << endl;
  ss << "  AARV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAARVT;
  ss << "  AERV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAERVT;
  ss << "  EERV = " << fixed << setw(14) << setprecision(4) << mysasparams->XEERVT;
  ss << endl;
  ss << "  AAVV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAAVVT;
  ss << "  AEVV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAEVVT;
  ss << "  EEVV = " << fixed << setw(14) << setprecision(4) << mysasparams->XEEVVT;
  ss << endl;
  
  ss << "Cases:    " << endl;
  ss << "  AARR = " << fixed << setw(14) << setprecision(4) << mysasparams->XAARRS;
  ss << "  AERR = " << fixed << setw(14) << setprecision(4) << mysasparams->XAERRS;
  ss << "  EERR = " << fixed << setw(14) << setprecision(4) << mysasparams->XEERRS;
  ss << endl;
  ss << "  AARV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAARVS;
  ss << "  AERV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAERVS;
  ss << "  EERV = " << fixed << setw(14) << setprecision(4) << mysasparams->XEERVS;
  ss << endl;
  ss << "  AAVV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAAVVS;
  ss << "  AEVV = " << fixed << setw(14) << setprecision(4) << mysasparams->XAEVVS;
  ss << "  EEVV = " << fixed << setw(14) << setprecision(4) << mysasparams->XEEVVS;
  ss << endl;
  */
  return ss.str();
}

void ScoreSAS::setOddsRatio(double r)  {
  if ( r <= RMIN )  {
    oddsratio = "small";
  }
  else if ( r >= RMAX )  {
    oddsratio = "large";
  }
  else  {
    ostringstream rstr;
    rstr << fixed << setw(10) << setprecision(4) << r;
    oddsratio = rstr.str();
  }
}









