#include "ScoreSASGC.h"


double ScoreSASGC::getSASEstimatedImputationR2(double *dosage, int *P, int *A, int nsamples, int anc)  {

  int n=0;
  double mean =0;
  double est_var = 0;
  double est_freq = 0;
  double est_allele_var =0;
  double rhat;

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || A[i] == 9 ) 
      continue;
    if(A[i] == anc ){
      mean += dosage[i];
      n++;
    }
  }
  est_freq = mean/(2*n);
  mean /= n;

  if(est_freq<0.01|| est_freq > 0.99){
    //fprintf(stderr,"potential overflow error in estimation of imputation error rate\n");
    return 1;
  }

  /*estimate variance of dosages*/
    est_allele_var = 2*(est_freq)*(1-est_freq);
  //est_allele_var = 2*(f)*(1-f);

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || A[i] == 9 || A[i] != anc) 
      continue;
    if(A[i] == anc ){
      double dose = dosage[i];
      est_var += ((dose-mean)*(dose-mean));
    }
  }
  est_var /= (n-1);

  rhat = (est_var / (est_allele_var));
  if(rhat<0.1){
    rhat=0.1;
  }
  if(rhat>1){
    rhat=1;
  }

  //fprintf(stderr,"anc %d mean %f freq %f est_var %f est_allele_var %f rhat %f\n",anc,mean,est_freq,est_var, est_allele_var, rhat);

  return rhat;
}
double ScoreSASGC::getSASEstimatedImputationR2(int *g, int *P, int *A, int nsamples, int anc)  {

  int n=0;
  double mean =0;
  double est_var = 0;
  double est_freq = 0;
  double est_allele_var =0;
  double rhat;

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || A[i] == 9 ) 
      continue;
    if(A[i] == anc ){
      mean += g[i];
      n++;
    }
  }
  est_freq = mean/(2*n);
  mean /= n;

  if(est_freq<0.01|| est_freq > 0.99){
    //fprintf(stderr,"potential overflow error in estimation of imputation error rate\n");
    return 1;
  }

  /*estimate variance of dosages*/
    est_allele_var = 2*(est_freq)*(1-est_freq);
  //est_allele_var = 2*(f)*(1-f);

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || A[i] == 9 || A[i] != anc) 
      continue;
    if(A[i] == anc ){
      double dose = (double)(g[i]);
      est_var += ((dose-mean)*(dose-mean));
    }
  }
  est_var /= (n-1);

  rhat = (est_var / (est_allele_var));
  if(rhat<0.1){
    rhat=0.1;
  }
  if(rhat>1){
    rhat=1;
  }

  //fprintf(stderr,"anc %d mean %f freq %f est_var %f est_allele_var %f rhat %f\n",anc,mean,est_freq,est_var, est_allele_var, rhat);

  return rhat;
}
ScoreSASGC::ScoreSASGC(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreSASGC::ScoreSASGC]  ";

  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile", &genoStream);
  dipGenoVec = new int[nsamples];

  ioh.openInput("phenofile", &phenoStream);
  dipPhenoVec  = new int[nsamples];

  globalAnc = new double[nsamples];

  genoDosages = new double[nsamples];
  geno_is_dosage=false;
  if( 0==ioh.getParam("dosages",&str) ){
    if(str.compare("YES")==0 ){
      geno_is_dosage = true;
    }
  }
  use_adjustment = false;
  if( 0==ioh.getParam("adjustment",&str) ){
    if(str.compare("YES")==0 ){
      use_adjustment = true;
    }
  }

  ioh.openInput("ancfile",&localAncStream);
  dipLocalAnc = new int[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  if ( ioh.getBoolParam("oddsratio") )  {
    ioh.openOutput("orfile",&oddsratioStream);
  }

  return;
}

ScoreSASGC::~ScoreSASGC(void)  {
  string str;

  // close file streams 
  if ( outputStream.is_open() )     {  outputStream.close();     }
  if ( oddsratioStream.is_open() )  {  oddsratioStream.close();  }
  if ( genoStream.is_open()   )     {  genoStream.close();       }
  if ( phenoStream.is_open()  )     {  phenoStream.close();      }
  if ( localAncStream.is_open() )  {  localAncStream.close();  }
  // mem: global anc stream already closed

  // free memory
  delete dipGenoVec;
  delete dipPhenoVec;
  delete globalAnc;
  //delete mysasparams_gc->TS;
  delete genoDosages;

  delete dipLocalAnc;
    //delete mysasparams_gc->AS;
  delete mysasparams_gc;

}

int ScoreSASGC::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreSASGC::evaluate]  ";

  bool output_to_file = ioh.getBoolParam("output2file"); 
  bool output_oddsratio = ioh.getBoolParam("oddsratio");

  try { ioh.readIntLine(&phenoStream, nsamples, &dipPhenoVec); }  
  catch ( Exception& e )  {
    errstream << "error reading phenotype" << endl;
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }
  computeMeanSdGamma(dipPhenoVec);

  str = ioh.getStringParam("thetafile");
  try { ioh.readOneColumnFile(str, nsamples, &globalAnc); }
  catch ( Exception& e )  {
    errstream << "error reading globalanc" << endl;
    errstream << e.getMessage();
    throw( Exception (errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    if(geno_is_dosage){
      try { ioh.readRealLine(&genoStream, nsamples, &genoDosages); }   
      catch ( Exception& e )  {
	errstream << "SNP # " << snpcount << endl;
	errstream << "error reading genotype" << endl;
	errstream << e.getMessage();
	throw( Exception(errstream.str()) );
      }
    }
    else  {
      try { ioh.readIntLine(&genoStream, nsamples, &dipGenoVec); }   
      catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading genotype" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
      }
    }

    try { ioh.readIntLine(&localAncStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }

    double s;
    try  {
      s = evalSAS();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }

    if ( output_oddsratio )  {
      oddsratioStream << fixed << setw(10) << 
         oddsratio << endl;
    }
  }

  return 0;
}




#define RMIN 0.1
#define RMAX 10.0
#define PAMIN 0.0001
#define PAMAX 0.9999
#define PEMIN 0.0001
#define PEMAX 0.9999

#define MAXBIGITER 100
#define MAXSMALLITER 40
#define CONVERGE 1e-6
#define SMALL 1e-4
#define DBG 0



double ScoreSASGC::evalSAS()  {
  ostringstream errstream;
  errstream << "[ScoreSASGC::evalSAS]  ";

  /* store initial values of parameters and compute null score */
  if(geno_is_dosage)  {
    init_sasparams_gc(genoDosages, dipPhenoVec, dipLocalAnc, globalAnc);
  }
  else  {
    init_sasparams_gc(dipGenoVec, dipPhenoVec, dipLocalAnc, globalAnc); 
  }

  /*simple minimization routine*/
  bool cont;
  double best_pa_null;
  double best_pe_null;
  double best_l_null;
  double best_pa_causal;
  double best_pe_causal;
  double best_r_causal;
  double best_l_causal;


#define EPS 1e-4
  double lr,ll;
  double cur_pa = (PAMIN+PAMAX)/2;
  double cur_pe = (PEMIN+PEMAX)/2;
  double cur_r = 1;
  bool bigloop = true;
  double  prev_l = -9999999;

  while(bigloop){
    /*maximize over pA*/
    cont = true;
    double xaleft = PAMIN;
    double xaright = PAMAX;
    while(cont){
      double amid = (xaleft+xaright)/2;
      lr = loglik(mysasparams_gc, amid+EPS, cur_pe, cur_r);
      ll = loglik(mysasparams_gc, amid-EPS, cur_pe, cur_r);   

      //      cerr<<xaleft<<" "<<xaright<<endl;
      //cerr<<ll<<" "<<lr<<endl;

      if(lr>ll){
	xaleft=amid;
      }else{
	if(ll>lr){
	  xaright=amid;
	}else{
	  cont = false;
	}
      }
      if(xaright - xaleft < 2*EPS){
	cont = false;
      }

    }
    cur_pa = (xaleft+xaright)/2;

    /*maximize over pE*/
    cont = true;
    double xeleft = PEMIN;
    double xeright = PEMAX;
    while(cont){
      double emid = (xeleft+xeright)/2;
      lr = loglik(mysasparams_gc,cur_pa, emid+EPS, cur_r);
      ll = loglik(mysasparams_gc,cur_pa, emid-EPS, cur_r);   
      if(lr>ll){
	xeleft=emid;
      }else{
	if(lr<ll){
	  xeright=emid;
	}else{
	  cont = false;
	}
      }
      if(xeright-xeleft<2*EPS){
	cont = false;
      }
    }

    cur_pe = (xeleft+xeright)/2;
    double l = loglik(mysasparams_gc, cur_pa, cur_pe, cur_r);

    if(l-prev_l < CONVERGE){
      bigloop = false;
    }
    prev_l = l;

  }
  best_pa_null = cur_pa;
  best_pe_null = cur_pe;
  best_l_null = prev_l;


  bigloop = true;
  prev_l = -9999999;
  while(bigloop){

    /*maximize over R*/
    double xrleft = RMIN;
    double xrright = RMAX;
    cont = true;
    while(cont){
      double rmid = (xrleft+xrright)/2;
      lr = loglik(mysasparams_gc, cur_pa, cur_pe, rmid+EPS);
      ll = loglik(mysasparams_gc, cur_pa, cur_pe, rmid-EPS);   
      if(lr>ll){
	xrleft=rmid;
      }else{
	if(lr<ll){
	  xrright=rmid;
	}else{
	  cont = false;
	}
      }
      if(xrright -xrleft <EPS){
	cont = 0;
      }
    }
    cur_r = (xrleft+xrright)/2;

    /*maximize over pA*/
    double xaleft = PAMIN;
    double xaright = PAMAX;
    cont = true;
    while(cont){
      double amid = (xaleft+xaright)/2;
      lr = loglik(mysasparams_gc, amid+EPS, cur_pe, cur_r);
      ll = loglik(mysasparams_gc, amid-EPS, cur_pe, cur_r);   
      if(lr>ll){
	xaleft=amid;
      }else
      if(lr<ll){
	xaright=amid;
      }else{
	cont = false;
      }
      if(xaright-xaleft < 2*EPS){
	cont = 0;
      }
    }
    cur_pa = (xaleft+xaright)/2;
    /*maximize over pE*/
    double xeleft = PEMIN;
    double xeright = PEMAX;
    cont = true;
    while(cont){
      double emid = (xeleft+xeright)/2;
      lr = loglik(mysasparams_gc, cur_pa, emid+EPS, cur_r);
      ll = loglik(mysasparams_gc, cur_pa, emid-EPS, cur_r);   
      if(lr>ll){
	xeleft=emid;
      }else
      if(lr<ll){
	xeright=emid;
      }else{
	cont = false;
      }
      //printf("%f %f\n",xeleft,xeright);
      if(xeright-xeleft < 2*EPS){
	cont = 0;
      }
    }
    cur_pe = (xeleft+xeright)/2;

    double l = loglik(mysasparams_gc, cur_pa, cur_pe, cur_r);
    
    if(l-prev_l < CONVERGE){
      bigloop = false;
    }
    prev_l = l;
  }
  best_pa_causal = cur_pa;
  best_pe_causal = cur_pe;
  best_l_causal = prev_l;
  best_r_causal = cur_r;


  double score = best_l_causal-best_l_null;
  if(score<0){
    score = 0;
  }

  // fprintf(stderr,"NULL %f %f %f %f, CAUSAL %f %f %f %f, score %f\n",best_pa_null,best_pe_null,1.0,best_l_null,
  //            best_pa_causal,best_pe_causal,best_r_causal,best_l_causal,score);

  return (score);

}


/* ------------------------------------------------------------------- */
void ScoreSASGC::init_sasparams_gc(int *G, int *P, int *A, double *T)  {

  sasparams_gc_t *p = new sasparams_gc_t;

  p->XAARRT = p->XAARVT = p->XAAVVT = p->XAARRS = p->XAARVS = p->XAAVVS =
  p->XEERRT = p->XEERVT = p->XEEVVT = p->XEERRS = p->XEERVS = p->XEEVVS =
  p->XAERRT = p->XAERVT = p->XAEVVT = p->XAERRS = p->XAERVS = p->XAEVVS = 0.0;

  p->XAVT = p->XART = p->XAVS = p->XARS = 0.0;
  p->XEVT = p->XERT = p->XEVS = p->XERS = 0.0;

  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || G[i] == 9 || A[i] == 9 ) 
      continue;

    int g = G[i];
    int a = A[i];

    if ( P[i] == 0 )  {              /* Controls */
      if ( g == 0 )  {               /* VV */
        if ( a == 0 )  {             /* AA */
	  p->XAAVVT += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAEVVT += 1;
	}
	else  {                      /* EE */
	  p->XEEVVT += 1;
	}
      }
      else if ( g == 1 )  {          /* RV      */
        if ( a == 0 )  {             /* AA */
	  p->XAARVT += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERVT += 1;
	}
	else  {                      /* EE */
	  p->XEERVT += 1;
	}
      }
      else  {                        /* RR      */
        if ( a == 0 )  {             /* AA */
	  p->XAARRT += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERRT += 1;
	}
	else  {                      /* EE */
	  p->XEERRT += 1;
	}
      }
    }
    else  {                          /* Cases    */
      if ( g == 0 )  {               /* VV      */
        if ( a == 0 )  {             /* AA */
	  p->XAAVVS += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAEVVS += 1;
	}
	else  {                      /* EE */
	  p->XEEVVS += 1;
	}
      }
      else if ( g == 1 )  {          /* RV      */
        if ( a == 0 )  {             /* AA */
	  p->XAARVS += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERVS += 1;
	}
	else  {                      /* EE */
	  p->XEERVS += 1;
	}
      }
      else  {                        /* RR      */
        if ( a == 0 )  {             /* AA */
	  p->XAARRS += 1;
	}
	else if ( a == 1 )  {        /* AE */
	  p->XAERRS += 1;
	}
	else  {                      /* EE */
	  p->XEERRS += 1;
	}
      }
    }
  }


  /* pack Cases into arrays */
  p->AS = new int[nsamples];
  p->TS = new double[nsamples]; 

  p->nt = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 0 || P[i] == 9 || A[i] == 9 || G[i] == 9)  
      continue;
    
    p->TS[p->nt] = T[i];
    p->AS[p->nt] = A[i];
    p->nt++;
  }


  p->XART = (2*p->XAARRT + p->XAARVT) + 0.5 * (2*p->XAERRT + p->XAERVT);
  p->XARS = (2*p->XAARRS + p->XAARVS) + 0.5 * (2*p->XAERRS + p->XAERVS);
  p->XERT = (2*p->XEERRT + p->XEERVT) + 0.5 * (2*p->XAERRT + p->XAERVT);
  p->XERS = (2*p->XEERRS + p->XEERVS) + 0.5 * (2*p->XAERRS + p->XAERVS);


  p->XAVT = (2*p->XAAVVT + p->XAARVT) + 0.5  * (2*p->XAEVVT + p->XAERVT);
  p->XAVS = (2*p->XAAVVS + p->XAARVS) + 0.5 * (2*p->XAEVVS + p->XAERVS);
  p->XEVT = (2*p->XEEVVT + p->XEERVT) + 0.5 * (2*p->XAEVVT + p->XAERVT);
  p->XEVS = (2*p->XEEVVS + p->XEERVS) + 0.5 * (2*p->XAEVVS + p->XAERVS);

  for(int x =0; x < 1; x++){
    //  fprintf(stderr,"%f %f %f %f %f %f %f %f\n",p->XART,p->XARS,p->XERT,p->XERS,p->XAVT,p->XAVS,p->XEVT,p->XEVS);
    
    double estpa = 0.001 + p->XART/(p->XART+p->XAVT);
    double estpe = 0.001 + p->XERT/(p->XERT+p->XEVT);
    double estcpa = 0.001 + p->XARS/(p->XARS+p->XAVS);
    double estcpe = 0.001 + p->XERS/(p->XERS+p->XEVS);

    //fprintf(stderr,"%f %f %f %f (%f %f)\n",estpa,estpe,estcpa,estcpe, estpa/(estpa+estpe),estcpa/(estcpa+estcpe));
    
    p->XART = (2*p->XAARRT + p->XAARVT) + (estpa/(estpa+estpe)) * (2*p->XAERRT + p->XAERVT);
    p->XARS = (2*p->XAARRS + p->XAARVS) + (estcpa/(estcpa+estcpe)) * (2*p->XAERRS + p->XAERVS);
    p->XERT = (2*p->XEERRT + p->XEERVT) + (estpe/(estpa+estpe)) * (2*p->XAERRT + p->XAERVT);
    p->XERS = (2*p->XEERRS + p->XEERVS) + (estcpe/(estcpa+estcpe)) * (2*p->XAERRS + p->XAERVS);
    
    
    p->XAVT = (2*p->XAAVVT + p->XAARVT) + ((1-estpa)/(2-estpa-estpe))  * (2*p->XAEVVT + p->XAERVT);
    p->XAVS = (2*p->XAAVVS + p->XAARVS) + ((1-estcpa)/(2-estcpa-estcpe)) * (2*p->XAEVVS + p->XAERVS);
    p->XEVT = (2*p->XEEVVT + p->XEERVT) + ((1-estpe)/(2-estpa-estpe)) * (2*p->XAEVVT + p->XAERVT);
    p->XEVS = (2*p->XEEVVS + p->XEERVS) + ((1-estcpe)/(2-estcpa-estcpe)) * (2*p->XAEVVS + p->XAERVS);
    
    //fprintf(stderr,"%f %f %f %f %f %f %f %f\n",p->XART,p->XARS,p->XERT,p->XERS,p->XAVT,p->XAVS,p->XEVT,p->XEVS);
  }
  if(p->XEVT<1){p->XEVT = 1;}
  if(p->XEVS<1){p->XEVS = 1;}

  double numE = (  p->XERT );
  double denE = (  p->XERT + p->XEVT );
  
  double numA = (  p->XART );
  double denA = (  p->XART + p->XAVT );
  p->impPa = ( p->XARS + p->XART) / (p->XARS+p->XART+p->XAVS+p->XAVT);
  p->impPe = ( p->XERS + p->XERT) / (p->XERS+p->XERT+p->XEVS+p->XEVT);


  p->rhatA = 1;
  p->rhatE = 1;
  if(use_adjustment){
    p->rhatA = getSASEstimatedImputationR2(G, P, A, nsamples, 0);
    p->rhatE = getSASEstimatedImputationR2(G, P, A, nsamples, 2);
  }


  p->R = 1.0;
  p->pE = numE/denE;
  p->pA = numA/denA;
  p->meanGamma = meanGamma;
  p->sdGamma = sdGamma;
  mysasparams_gc = p;
}

/* ------------------------------------------------------------------- */
void ScoreSASGC::init_sasparams_gc(double *D, int *P, int *A, double *T)  {

  sasparams_gc_t *p = new sasparams_gc_t;
  int i;

  p->XAARRT = p->XAARVT = p->XAAVVT = p->XAARRS = p->XAARVS = p->XAAVVS =
  p->XEERRT = p->XEERVT = p->XEEVVT = p->XEERRS = p->XEERVS = p->XEEVVS =
  p->XAERRT = p->XAERVT = p->XAEVVT = p->XAERRS = p->XAERVS = p->XAEVVS = 0.0;

  p->XAVT = p->XART = p->XAVS = p->XARS = 0.0;
  p->XEVT = p->XERT = p->XEVS = p->XERS = 0.0;

  for(i=0;i<nsamples;i++)  {
    if ( P[i] == 9 || A[i] == 9 ) 
      continue;

    if(D[i] <0 || D[i] > 2)
      continue;

    int a = A[i];
    double d = D[i];


    if(P[i] == 0){
      
      if(a == 0){
	p->XAVT += (2-d);
	p->XART += (d);
      }
      if(a == 2){
	p->XEVT += (2-d);
	p->XERT += (d);
      }
      if(a == 1){
	p->XAVT += (2-d)/2;
	p->XART += (d)/2;
	p->XEVT += (2-d)/2;
	p->XERT += (d)/2;
      }
    }else{
      if(a == 0){
	p->XAVS += (2-d);
	p->XARS += (d);
      }
      if(a == 2){
	p->XEVS += (2-d);
	p->XERS += (d);
      }
      if(a == 1){
	p->XAVS += (2-d)/2;
	p->XARS += (d)/2;
	p->XEVS += (2-d)/2;
	p->XERS += (d)/2;
      }
    }
  }


  /* pack Cases into arrays */

  p->TS = new double[nsamples];
  p->AS = new int[nsamples];
  p->nt = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( P[i] == 0 || P[i] == 9 || A[i] == 9)  
      continue;
    
    p->TS[p->nt] = T[i];
    p->AS[p->nt] = A[i];
    p->nt++;
  }

  double numE = (  p->XERT );
  double denE = (  p->XERT + p->XEVT );
  
  double numA = (  p->XART );
  double denA = (  p->XART + p->XAVT );

  p->rhatA = 1;
  p->rhatE = 1;
  if(use_adjustment){
    p->rhatA = getSASEstimatedImputationR2(D, P, A, nsamples, 0);
    p->rhatE = getSASEstimatedImputationR2(D, P, A, nsamples, 2);
  }

  p->impPa = ( p->XARS + p->XART) / (p->XARS+p->XART+p->XAVS+p->XAVT);
  p->impPe = ( p->XERS + p->XERT) / (p->XERS+p->XERT+p->XEVS+p->XEVT);

  p->R = 1.0;
  p->pE = numE/denE;
  p->pA = numA/denA;
  p->meanGamma = meanGamma;
  p->sdGamma = sdGamma;

  mysasparams_gc = p;
}

void ScoreSASGC::computeMeanSdGamma(int *phenoVec)  {
  ifstream tempStream;
  ostringstream errstream;
  double *gamma=new double[nsnps];

  ioh.openInput("ancfile",&tempStream);
  for(int snpcount=0;snpcount<nsnps;snpcount++) {
    try { ioh.readIntLine(&tempStream, nsamples, &dipLocalAnc); }  
    catch ( Exception& e ) {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error reading diploid localanc" << endl;
      errstream << e.getMessage();
      throw( Exception(errstream.str()) );
    }
    gamma[snpcount] = 0;
    int ncase = 0;
    for(int i=0;i<nsamples;i++)  {
      if ( dipPhenoVec[i] == 1 )  {
	gamma[snpcount] += dipLocalAnc[i];
	ncase += 1;
      }
    }
    gamma[snpcount] /= (2*ncase);
    //cerr<<snpcount  <<" " <<  gamma[snpcount] << endl;
  }
  tempStream.close();
  double sum =0;
  for(int snpcount=0;snpcount<nsnps;snpcount++) {
    sum+=gamma[snpcount];
  }
  meanGamma = sum/nsnps;
  sdGamma = 0;
  for(int snpcount=0;snpcount<nsnps;snpcount++) {
    sdGamma += pow((gamma[snpcount]-meanGamma),2);
  }

  sdGamma = sqrt (sdGamma/(nsnps-1));
  delete gamma;
  //cerr<<meanGamma<< " "<<sdGamma<<endl;
  //  getchar();


}



/* ------------------------------------------------------------------- */
double ScoreSASGC::loglik(sasparams_gc_t *p, double pA, double pE, double R )  {
  int nt =  p->nt;
  double meanGamma = p->meanGamma;
  double sdGamma = p->sdGamma;

  double Ra = (1+(R-1)*(p->rhatA + pA*(1-p->rhatA)))
    / (1+(R-1)*pA*(1-p->rhatA));
  double Re = (1+(R-1)*(p->rhatE + p->pE*(1-p->rhatE)))
    / (1+(R-1)*pE*(1-p->rhatE));

  int *AS;
  AS = p->AS;

  double casepA, casepE, result, resultD;
  int i;

  casepA = Ra*pA / (1.0 + (Ra-1.0)*pA);
  casepE = Re*pE / (1.0 + (Re-1.0)*pE);
  result = 0.0;
  resultD = 0.0;

  result += p->XERT *log(pE) + p->XEVT *log(1-pE) + p->XART * log (pA) + p->XAVT *log(1-pA);
  result += p->XERS *log(casepE) + p->XEVS *log(1-casepE) + p->XARS * log (casepA) + p->XAVS *log(1-casepA);

  //admixture component
  double gammai=0;
  for(i=0;i<nt;i++)  {
    gammai+= AS[i];
  }
  gammai /= (2*nt);
  
  //compute new mean as function of rho and sdGamma meanGamma;
#if 1
  double omegaR = ( pE*R + (1-pE) )/( pA*R + (1-pA) );
  double meanGammaR = (meanGamma * omegaR) / (omegaR*meanGamma + 1 - meanGamma);
  double admL = -0.5 * pow( (gammai - meanGammaR),2) / (sdGamma*sdGamma);
  //cerr<<R<<" "<<omegaR << " "<< meanGamma<< " " <<meanGammaR <<" " << admL<<endl;getchar();
  result += admL;
#endif
#if 0
  //old version
  double sum = 0.0;
  double rho = ( pE*R + (1-pE) )/( pA*R + (1-pA) );
  for(i=0;i<nt;i++)  {
    double p2 = rho*rho*T[i]*T[i];
    double p1 = 2*rho*T[i]*(1-T[i]);
    double p0 = (1-T[i])*(1-T[i]);
    double psum = p0+p1+p2;
    if ( p->local_anc_is_prob )  {
      if ( A2[i] > 1e-8 )  
        sum += A2[i]*log( p2/psum );
      if ( A1[i] > 1e-8 ) 
        sum += A1[i]*log( p1/psum );
      if ( A0[i] > 1e-8 )
        sum += A0[i]*log( p0/psum );
    }
    else  {
      if ( AS[i] == 2 )  
        sum += log( p2/psum );
      else if ( AS[i] == 1 ) 
        sum += log( p1/psum );
      if ( AS[i] == 0 )
        sum += log( p0/psum );
    }
  }
  result += sum;
#endif


  return ( (2*result) );    /* maximize by minimizing negative */

}

/* ------------------------------------------------------------------- */
#define BINSEARCH_MAXITER 100
int ScoreSASGC::binsearch(double(*f)(double, sasparams_gc_t *),
  sasparams_gc_t *p, double xmin, double xmax, double *xout)  {
  ostringstream errstream;
  errstream << "[ScoreSASGC::binsearch]  ";

  double xleft  = xmin;
  double xright = xmax;
  double fleft  = f(xleft, p);
  double fright = f(xright, p);
  int itcount = 0;

  if ( fleft*fright > 0 )  {
    errstream << "no minimum on search interval";
    throw( Exception(errstream.str()) );
  }

  while ( (xright-xleft) > SMALL )  {

    double xmid = 0.5*(xleft+xright);
    double fmid = f(xmid, p);

    if ( fmid*fleft < 0 )  {
      xright = xmid;
      fright = fmid;
    }
    else if ( fmid*fright < 0 )  {
      xleft = xmid;
      fleft = fmid;
    }
    else  {
      return -1;
    }

    if ( itcount++ > BINSEARCH_MAXITER )  {
      errstream << "did not converge in " << (itcount-1) << "iterations";
      throw( Exception(errstream.str()) );
    }
  }

  *xout = 0.5*(xleft+xright);
  return 0;
}




/* ------------------------------------------------------------------- */
string ScoreSASGC::errorString(void)  {
  ostringstream ss;
  ss << "R = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->R;
  ss << "  pA = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->pA;
  ss << "  pE = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->pE;
  ss << endl;
  ss << "Controls: " << endl;
  ss << " AR = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XART ;
  ss << " ER = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XERT ;
  ss << " AV = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XAVT ;
  ss << " EV = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XEVT ;
  ss <<endl;
  ss << "Controls: " << endl;
  ss << " AR = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XARS ;
  ss << " ER = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XERS ;
  ss << " AV = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XAVS ;
  ss << " EV = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->XEVS ;
  ss <<endl;
  ss << "Imputation: " << endl;

  ss<< " RhatA = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->rhatA ;
  ss<< " RhatE = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->rhatE ;
  ss<< " fA = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->impPa ;
  ss<< " fE = "<< fixed << setw(14) << setprecision(4) << mysasparams_gc->impPe ;
  ss<<endl;

  /*

  ss << "  AARR = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAARRT;
  ss << "  AERR = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAERRT;
  ss << "  EERR = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XEERRT;
  ss << endl;
  ss << "  AARV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAARVT;
  ss << "  AERV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAERVT;
  ss << "  EERV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XEERVT;
  ss << endl;
  ss << "  AAVV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAAVVT;
  ss << "  AEVV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAEVVT;
  ss << "  EEVV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XEEVVT;
  ss << endl;
  
  ss << "Cases:    " << endl;
  ss << "  AARR = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAARRS;
  ss << "  AERR = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAERRS;
  ss << "  EERR = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XEERRS;
  ss << endl;
  ss << "  AARV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAARVS;
  ss << "  AERV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAERVS;
  ss << "  EERV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XEERVS;
  ss << endl;
  ss << "  AAVV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAAVVS;
  ss << "  AEVV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XAEVVS;
  ss << "  EEVV = " << fixed << setw(14) << setprecision(4) << mysasparams_gc->XEEVVS;
  ss << endl;
  */
  return ss.str();
}

void ScoreSASGC::setOddsRatio(double r)  {
  if ( r <= RMIN )  {
    oddsratio = "small";
  }
  else if ( r >= RMAX )  {
    oddsratio = "large";
  }
  else  {
    ostringstream rstr;
    rstr << fixed << setw(10) << setprecision(4) << r;
    oddsratio = rstr.str();
  }
}









