#include "ScoreZ2.h"

ScoreZ2::ScoreZ2(char *paramfile)  {
  string str, a;
  ostringstream errstream;
  errstream << "[ScoreZ2::ScoreZ2]  ";

  pheno_is_quantitative = 0;

  ioh.readParams(paramfile);

  nsamples = ioh.getPositiveIntParam("nsamples");
  nsnps = ioh.getPositiveIntParam("nsnps");

  ioh.openInput("genofile",&genoStream);
  dipGenoVec = new double[nsamples];

  ioh.openInput("phenofile",&phenoStream);
  dipPhenoVec  = new double[nsamples];
  globalAnc = new double[nsamples];

  if ( ioh.getBoolParam("output2file") )  {
    ioh.openOutput("outfile",&outputStream);
  }

  pheno_is_quantitative = ioh.getBoolParam("qpheno");
  ioh.getParam("genoformat",&genoform);

  if ( ioh.getBoolParam("effectsize") &&
       pheno_is_quantitative )  {
    ioh.openOutput("esfile",&effectsizeStream);
  }

  return;
}

void ScoreZ2::scoreIsQ(void)  {
  pheno_is_quantitative = true;
}

ScoreZ2::~ScoreZ2(void)  {
  string str;

  // close file streams 
  if ( outputStream.is_open() )     { outputStream.close();     }
  if ( effectsizeStream.is_open() ) { effectsizeStream.close(); }
  if ( genoStream.is_open()   )     { genoStream.close();       }
  if ( phenoStream.is_open()  )     { phenoStream.close();      }
  // mem: global anc stream already closed

  // free memory
  delete dipGenoVec;
  delete dipPhenoVec;
  delete globalAnc;

}

int ScoreZ2::evaluate()  {
  string str;
  ostringstream errstream;
  errstream << "[ScoreZ2::evaluate]  ";

  bool output_effectsize = ioh.getBoolParam("effectsize");

  int *tempIntPheno;
  if ( !pheno_is_quantitative )  {
    tempIntPheno = new int[nsamples];
  }

  ioh.getParam("output2file",&str);
  int output_to_file = ( !str.compare("YES") );

  try { 
    if ( pheno_is_quantitative )  {
      ioh.readRealLine(&phenoStream, nsamples, &dipPhenoVec); 
      for(int k=0;k<nsamples;k++)  {
        if ( fabs(dipPhenoVec[k]+1024.0) < 1e-6  )  {
          dipPhenoVec[k] = -1.0;  //  questionable:  revisit
        }
      }
    }
    else  {
      ioh.readIntLine(&phenoStream, nsamples, &tempIntPheno); 
      for(int k=0;k<nsamples;k++)  {
        if ( tempIntPheno[k] == 9 )  {
          dipPhenoVec[k] = -1.0;  //  questionable:  revisit
        }
        else  {
          dipPhenoVec[k] = (double) tempIntPheno[k];
        }
      }
    }
  }  
  catch ( Exception& e )  {
    errstream << "error reading phenotype" << endl;
    errstream << endl << e.getMessage();
    throw( Exception(errstream.str()) );
  }

  ioh.getParam("thetafile", &str);
  try { ioh.readOneColumnFile(str, nsamples, &globalAnc); }
  catch ( Exception& e )  {
    errstream << "error reading globalanc" << endl;
    errstream << e.getMessage();
    throw( Exception (errstream.str()) );
  }

  for(int snpcount=0;snpcount<nsnps;snpcount++)  {

    if( !genoform.compare("dosage") ){
      try { ioh.readRealLine(&genoStream, nsamples, &dipGenoVec); }   
      catch ( Exception& e )  {
	errstream << "SNP # " << snpcount << endl;
	errstream << "error reading genotype" << endl;
	errstream << e.getMessage();
	throw( Exception(errstream.str()) );
      }
    }
    else if ( !genoform.compare("prob") )  {
      double *tempVec = new double[2*nsamples];
      try { ioh.readRealLine(&genoStream, 2*nsamples, &tempVec); }   
      catch ( Exception& e )  {
	errstream << "SNP # " << snpcount << endl;
	errstream << "error reading genotype" << endl;
	errstream << e.getMessage();
	throw( Exception(errstream.str()) );
      }
      for(int i=0;i<nsamples;i++)  {
        double p2 = tempVec[2*i];
        double p1 = tempVec[2*i+1];
	dipGenoVec[i] = 2.0*p2 + p1;
        if ( p2 == 0.333 && p1 == 0.333 )  {
          dipGenoVec[i] = 9;
        }
      }
      delete tempVec;
    }
    else  {    // genoformat is "integer"
      int *tempVec = new int[nsamples];
      try { ioh.readIntLine(&genoStream, nsamples, &tempVec); }   
      catch ( Exception& e )  {
	errstream << "SNP # " << snpcount << endl;
	errstream << "error reading genotype" << endl;
	errstream << e.getMessage();
	throw( Exception(errstream.str()) );
      }
      for(int i = 0;i < nsamples;i++){
	dipGenoVec[i] = (double)tempVec[i]; 
      }
      delete tempVec;
    }

    double s;
    try  {
      s = evalZ2();
    }
    catch ( Exception& e )  {
      errstream << "SNP # " << snpcount << endl;
      errstream << "error computing score" << endl;
      errstream << e.getMessage();

      // print error data to standard error and try to go on to next SNP
      cerr << errstream.str();
      if ( output_to_file )  {   // place-holder in output file
        outputStream << "no-score" << endl;
      }
      continue;

    }

    if ( output_to_file )  {
      outputStream << fixed << setw(10) << setprecision(4) << s << endl;
    }
    else  {
      cout << fixed << setw(10) << setprecision(4) << s << endl;
    }

    if ( output_effectsize )  {
      effectsizeStream << fixed << setw(10) << effectsize << endl;
    }
  }

  if ( !pheno_is_quantitative )  {
    delete tempIntPheno;
  }

  return 0;
}


#define VERYSMALL 1e-8

/* ------------------------------------------------------------------- */
/*          Score Zero with regression                                 */
/* ------------------------------------------------------------------- */
/*  parameters : (in)                                                  */
/*       G - diploid genotype (0,1,2,9)                                */
/*       P - phenotype (0,1,9)                                         */
/*       T - global ancestry                                           */
/*  returns score or sets global errflag                               */
/* ------------------------------------------------------------------- */
double ScoreZ2::evalZ2(void)  {
  ostringstream errstream;
  errstream << "[ScoreZ2::evalZ2] ";

  /* mean-adjusted genotype */
  double *gamma = new double[nsamples];

  /* mean-adjusted phenotype */
  double *phi = new double[nsamples];

  /* mean-adjusted global ancestry */
  double *theta = new double[nsamples];

  /* missing data flag */
  int *baddata = new int[nsamples];

  double sumx, sumy, sumxx, sumxy, sumyy, numer, denom1, denom2, corr;
  double gdott, pdott, tdott;
  double sum, mean;
  int count;

  /* find bad data and flag those as unused */
  for(int i=0;i<nsamples;i++)  {
    baddata[i] = ( dipGenoVec[i] == 9 || 
                   fabs(dipPhenoVec[i]+1.0) < 1e-6  ||
		   dipGenoVec[i] > 2 ||
		   dipGenoVec[i] < 0 
		   );
  }

  /* mean-adjust theta */
  count = 0;
  sum = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    count++;
    sum += globalAnc[i];
  }
  mean = sum/count;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    theta[i] = globalAnc[i] - mean;
  }

  /* mean-adjust G */
  count = 0;
  sum = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    count++;
    sum += (double) dipGenoVec[i];
  }
  mean = sum/count;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    gamma[i] = (double) dipGenoVec[i] - mean;
  }

  /* mean-adjust P */
  count = 0;
  sum = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    count++;
    sum += (double) dipPhenoVec[i];
  }
  mean = sum/count;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    phi[i] = dipPhenoVec[i] - mean;
  }

  /* regression */
  pdott = 0.0;
  tdott = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    pdott += phi[i]*theta[i];
    tdott += theta[i]*theta[i];
  }
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    phi[i] -= (pdott/tdott)*theta[i];
  }

  gdott = 0.0;
  tdott = 0.0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    gdott += gamma[i]*theta[i];
    tdott += theta[i]*theta[i];
  } 
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    gamma[i] -= (gdott/tdott)*theta[i];
  }

  /* correlation coefficient */
  sumx = sumy = sumxx = sumxy = sumyy = 0.0;
  count = 0;
  for(int i=0;i<nsamples;i++)  {
    if ( baddata[i] ) continue;
    sumx  += gamma[i];
    sumxx += gamma[i]*gamma[i];
    sumy  += phi[i];
    sumyy += phi[i]*phi[i];
    sumxy += gamma[i]*phi[i];
    count++;
  }

  delete gamma;
  delete phi; 
  delete theta;
  delete baddata;

  numer = sumxy/count - (sumx/count)*(sumy/count);
  denom1 = (sumxx/count - (sumx/count)*(sumx/count));
  denom2 = (sumyy/count - (sumy/count)*(sumy/count));

  if ( denom1 < VERYSMALL || denom2 < VERYSMALL )  {
    errstream << "zero standard deviation";
    throw(Exception(errstream.str()));
  }
  else  {
    corr = numer/sqrt(denom1*denom2);
    setEffectSize(sumxy/sumxx);
  }
  return((count-2)*corr*corr);

}

void ScoreZ2::setEffectSize(double eff)  {
  ostringstream rstr;
  rstr << fixed << setw(10) << setprecision(4) 
    << eff;
  effectsize = rstr.str();
}
